/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.typing;

import java.util.List;

import org.xml.sax.SAXException;
import org.xquark.schema.AttributeDeclaration;
import org.xquark.schema.ChoiceModelGroup;
import org.xquark.schema.ComplexType;
import org.xquark.schema.ElementDeclaration;
import org.xquark.schema.ModelGroup;
import org.xquark.schema.Particle;
import org.xquark.schema.Schema;
import org.xquark.schema.SchemaException;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.SchemaScope;
import org.xquark.schema.SequenceModelGroup;
import org.xquark.schema.SimpleType;
import org.xquark.schema.Type;
import org.xquark.schema.Wildcard;
import org.xquark.xquery.parser.QName;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryExpression;

public class ElementTools {

	private static final String RCSRevision = "$Revision: 1.4 $";

	private static final String RCSName = "$Name:  $";

	private byte index = 0;

	private static final String TEMP_NAME = "TEMP_NAME";

	private static final String STAR = "*";

	public static final String SC_ANYTYPE = "anyType";

	public static final String SC_STRING = "string";

	private static final String XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";

	private int countSimpleType = 0;

	private int countAttribute = 0;

	private int countSimpleList = 0;

	private int size = 0;

	private boolean mixed = false;

	private SimpleType simpleType = null;

	private XQueryExpression name = null;

	private Particle wildcard = null;

	Schema schema = null;

	SchemaManager schemamanager = null;

	public ElementTools(int size, XQueryExpression name, Schema schema, SchemaManager schemamanager) {
		this.size = size;
		this.name = name;
		this.schema = schema;
		this.schemamanager = schemamanager;
	}

	// ******************************************
	// FOR ACCESS TO PRIVATE/PROTECTED ATTRIBUTES
	// ******************************************

	public int getCountSimpleType() {
		return countSimpleType;
	}

	// public void setCountSimpleType(int countSimpleType) { this.countSimpleType = countSimpleType; }

	public int getCountAttribute() {
		return countAttribute;
	}

	// public void setCountAttribute(int countAttribute) { this.countAttribute = countAttribute; }

	public int getCountSimpleList() {
		return countSimpleList;
	}

	// public void setCountSimpleList(int countSimpleList) { this.countSimpleList = countSimpleList; }

	public int getSize() {
		return size;
	}

	public void setSize(int size) {
		this.size = size;
	}

	public boolean getMixed() {
		return mixed;
	}

	public void setMixed(boolean mixed) {
		this.mixed = mixed;
	}

	public SimpleType getSimpleType() {
		return simpleType;
	}

	public void setSimpleType(SimpleType simpleType) {
		this.simpleType = simpleType;
	}

	public XQueryExpression getName() {
		return name;
	}

	public void setName(XQueryExpression name) {
		this.name = name;
	}

	public void addWildCard(Particle wildcard) {
		this.wildcard = wildcard;
	}

	public Particle getWildCard() {
		return wildcard;
	}

	// Increment methods

	public void incSize() {
		size++;
	}

	public void incCountSimpleType() {
		countSimpleType++;
	}

	public void incCountAttribute() {
		countAttribute++;
	}

	public void incCountSimpleList() {
		countSimpleList++;
	}

	public void buildQTypeAtom(QType qtype) {
		setSimpleType(qtype.getSimpleType());
		setMixed(true);
		incCountSimpleType();
		incCountSimpleList();
	}

	public void buildQTypeText(QType qtype) {
		setSimpleType(qtype.getSimpleType());
		setMixed(true);
		incCountSimpleType();
		incCountSimpleList();
	}

	private String getTemporaryName() {
		return TEMP_NAME + index++;
	}

	public void buildQTypeAttribute(QTypeAttribute qtype, ComplexType complexType) throws XQueryException {
		String name = null;
		XQueryExpression qname = qtype.getName();
		if (qname instanceof QName)
			name = ((QName) qtype.getName()).getName();
		else
			name = getTemporaryName();
		// add LARS 05/07/04
		if (name.equals("xmlns"))
			return;
		incCountAttribute();
		AttributeDeclaration attribute = new AttributeDeclaration(schema, name, complexType);
		attribute.setType(qtype.getSimpleType());
		try {
			complexType.register(attribute);
		} catch (SchemaException saxe) {
			throw new TypeException("Error in register(ElementDeclaration): " + saxe.getMessage());
		}
	}

	public void buildQTypeElement(QType qtype, ComplexType complexType, ModelGroup model, boolean hasOne,
			boolean unbounded) throws XQueryException {
		String name = null;
		String uri = null;
		Schema tmpschema = schema;
		try {
			QName qname = (QName) ((QTypeElement) qtype).getName();
			uri = qname.getNameSpace();
			if (uri != null && !uri.equals(STAR)) {
				tmpschema = schemamanager.getSchema(uri);
				if (tmpschema == null) {
					tmpschema = new Schema(uri, null, schemamanager);
					try {
						schemamanager.putSchema(tmpschema);
					} catch (SAXException saxe) {
						throw new TypeException("Could not add schema to manager");
					}
				}
			}
			name = qname.getName();
		} catch (ClassCastException cce) {
			name = "TagDefineInEvaluation";
		}
		Particle particle = null;
		if (!name.equals(STAR)) {
			ElementDeclaration element = null;
			// test if element already exists
			element = complexType.getElementDeclaration(uri, name);
			if (element == null) {
				SchemaScope scope = null;
				if (uri != null && !uri.equals(STAR)) {
					scope = tmpschema;
					element = new ElementDeclaration(tmpschema, name, scope);
					element.setQualified(true);
				} else {
					scope = complexType;
					element = new ElementDeclaration(tmpschema, name, scope);
				}
				element.setType(qtype.getType());
				try {
					if (uri != null && !uri.equals(STAR))
						complexType.register(element);
					else
						scope.register(element);
				} catch (SchemaException se) {
					throw new XQueryException(se.getMessage());
				}
			} else {
				element.setType(tmpschema.getType(SC_ANYTYPE));
				/*
				 * Type elementType = element.getType(); if (elementType.getValueType() != null)
				 */
			}
			particle = new Particle(element);
		} else { // name = STAR
			// TODO? make a choice of different wildcards?
			particle = getWildCard();
			if (particle == null) {
				particle = new Particle(new Wildcard(tmpschema, true));
				addWildCard(particle);
			}
		}
		if (!hasOne)
			particle.setMinOccurs(0);
		if (unbounded)
			particle.setMaxOccursUnbounded();
		model.add(particle);
	}

	public QTypeElement buildQTypeSequence(QType qtype, ComplexType complexType, SequenceModelGroup sequenceModel)
			throws XQueryException {
		List list = qtype.getList();
		if (list == null)
			return null;
		int size = list.size();
		QTypeElement qte = null;
		for (int i = 0; i < size; i++) {
			QType stype = (QType) list.get(i);
			int subclass = stype.getSubClass();
			if (subclass == QType.ATOM) {
				// case of <tag> [simpleType] </tag>, like <a> 23 </a>
				if (size == 1 && getSize() == 1)
					return new QTypeElement(getName(), qtype.getSimpleType() == null ? getType(SC_ANYTYPE) : qtype
							.getSimpleType());
				else
					buildQTypeAtom(stype);
			} else if (subclass == QType.TEXT) {
				buildQTypeText(stype);
			} else if (subclass == QType.ATTRIBUTE) {
				buildQTypeAttribute((QTypeAttribute) stype, complexType);
			} else if (subclass == QType.ELEMENT) {
				buildQTypeElement(stype, complexType, sequenceModel, stype.getMinOccurence() == 1, stype.isMultiple());
			} else if (subclass == QType.UNION) {
				buildQTypeUnion(stype, complexType, sequenceModel);
			} else if (subclass == QType.SEQUENCE) {
				qte = buildQTypeSequence(stype, complexType, sequenceModel);
				if (qte != null)
					return qte;
			}

		}
		return null;
	}

	public void buildQTypeUnion(QType qtype, ComplexType complexType, SequenceModelGroup sequenceModel)
			throws XQueryException {
		List list = qtype.getList();
		ChoiceModelGroup choiceModel = null;
		for (int i = 0; i < list.size(); i++) {
			if (i != 0)
				incSize();
			QType utype = (QType) list.get(i);
			int subclass = utype.getSubClass();
			boolean atom = false;
			switch (subclass) {
			case QType.ATOM:
				if (!atom) {
					buildQTypeAtom(utype);
					atom = true;
				}
				break;
			case QType.ATTRIBUTE:
				buildQTypeAttribute((QTypeAttribute) utype, complexType);
				break;
			case QType.ELEMENT:
				if (choiceModel == null)
					choiceModel = new ChoiceModelGroup();
				buildQTypeElement(utype, complexType, choiceModel, utype.getMinOccurence() == 1, utype.isMultiple());
				break;
			case QType.SEQUENCE:
				buildQTypeSequence(utype, complexType, sequenceModel);
				break;
			case QType.TEXT:
				buildQTypeText(utype);
				break;
			case QType.PROCESSING:
			case QType.COMMENT:
			case QType.DOCUMENT:
			case QType.UNION:
				throw new TypeException("Could not type, incorrect element for element content.");
			}
			/*
			 * if (subclass == QType.ATOM) { if (!atom) { buildQTypeAtom(eTools, utype); atom = true; } } else if
			 * (subclass == QType.TEXT) { buildQTypeText(eTools, utype); } else if (subclass == QType.ATTRIBUTE) {
			 * buildQTypeAttribute(eTools, (QTypeAttribute) utype, complexType); } else if (subclass == QType.ELEMENT) {
			 * if (choiceModel == null) choiceModel = new ChoiceModelGroup(); buildQTypeElement(eTools, utype,
			 * complexType, choiceModel, utype.getMinOccurence() == 1, utype .isMultiple()); } else if (subclass ==
			 * QType.SEQUENCE) { this.buildQTypeSequence(eTools, utype, complexType, sequenceModel); } else if (subclass ==
			 * QType.DOCUMENT) { } else throw new TypeException( "We have no specification for this case (NYI),
			 * definition: union { (atom | element | attribute)+ }.");
			 */
		}
		if (choiceModel != null)
			sequenceModel.add(new Particle(choiceModel));
	}

	private Type getType(String typeStr) {
		return schemamanager.getType(XML_SCHEMA, typeStr);
	}

	// ***********************
	// STUFF JUST FOR PRINTING
	// ***********************

	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("*** ElementTools ***\n");
		sb.append("countSimpleType: " + countSimpleType + "\n");
		sb.append("countAttribute: " + countAttribute + "\n");
		sb.append("countSimpleList: " + countSimpleList + "\n");
		sb.append("size: " + size + "\n");
		sb.append("mixed: " + mixed + "\n");
		sb.append("********************");
		return sb.toString();
	}
}
