/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.typing;

import org.xquark.schema.SimpleType;
import org.xquark.xquery.parser.XQueryException;

public class QTypeAtom extends QTypeItem implements QTypeVisitable {
	private static final String RCSRevision = "$Revision: 1.11 $";

	private static final String RCSName = "$Name:  $";

	public QTypeAtom(SimpleType simpleType) {
		setType(simpleType);
		subclass = ATOM;
		occurence = OCC_1_1;
	}

	public QTypeAtom(SimpleType simpleType, byte occurence) {
		setType(simpleType);
		subclass = ATOM;
		this.occurence = occurence;
	}

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(QTypeVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// get/set
	// public SimpleType getSimpleType() {
	// return (SimpleType)type;
	// }

	// equals

	public boolean equals(Object qtype) {
		if (!(qtype instanceof QTypeAtom))
			return false;
		// if (((QTypeAtom) qtype).getOccurence() != occurence)
		// return false;
		return compare(type, ((QTypeAtom) qtype).getSimpleType());
	}

	// tells whether qtype is numeric
	public boolean isNumeric() {
		if (type == null)
			return false;
		else
			return isNumeric(type);
	}

	// tells whether qtype is boolean
	public boolean isBoolean() {
		if (type == null)
			return false;
		else
			return isBoolean(type);
	}

	// tells whether qtype is integer
	public boolean isInteger() {
		if (type == null)
			return false;
		else
			return isInteger(type);
	}

	// tells whether qtype is string
	public boolean isString() {
		if (type == null)
			return false;
		else
			return isString(type);
	}

	// tells whether qtype is an atom
	public boolean isAtom() {
		return true;
	}

	// // tells whether qtype can react to text()
	// public boolean isText() {
	// return false;
	// }

	// tells whether qtype is a date or collection of dates
	public boolean isDate() {
		if (type == null)
			return false;
		else
			return isDate(type);
	}

    // tells whether qtype is a date or collection of dates
    public boolean isDateTime() {
        if (type == null)
            return false;
        else
            return isDateTime(type);
    }

    // tells whether qtype is a date or collection of dates
    public boolean isDuration() {
        if (type == null)
            return false;
        else
            return isDuration(type);
    }

    // tells whether qtype is a date or collection of dates
    public boolean isTime() {
        if (type == null)
            return false;
        else
            return isTime(type);
    }

	// // tells whether qtype is a attribte or collection or sequence
	// public boolean isAttribute() {
	// return false;
	// }

	// // tells whether qtype is a ID, IDREF or IDREFS type
	// public boolean isIDAttribute() {
	// return false;
	// }

	// // tells whether qtype is a node
	// public boolean isNode() {
	// return false;
	// }

	// // tells whether qtype is a node with simple type
	// public boolean isSimpleTypeNode() {
	// return false;
	// }

	// // tells whether qtype is a collection
	// public boolean isCollection() {
	// return false;
	// }

	// // tells whether qtype is a document
	// public boolean isDocument() {
	// return false;
	// }

	// // tells whether qtype is multiple
	// public boolean isMultiple() {
	// return false;
	// }

	// tells whether qtype is atomic (atom or element with simple content
	public boolean isAtomic() {
		if (type == null)
			return false;
		else
			return !isMultiple();
	}

	// // tells whether qtype is mixed
	// public boolean isMixed() {
	// return false;
	// }

	// // tells whether qtype is element
	// public boolean isElement() {
	// return false;
	// }

	// tells whether qtype is qname
	public boolean isQName() {
		if (type == null)
			return false;
		else
			return isQName(type);
	}

	public boolean canBeComparedTo(QType compType) {
		if (compType == null)
			return false;
		if (!(compType instanceof QTypeAtom))
			return compType.canBeComparedTo(this);
		if (type == null)
			return true;
		if (((QTypeAtom) compType).getSimpleType() == null)
			return true;
		return canBeCompared(((SimpleType) type).getPrimitive(), ((QTypeAtom) compType).getSimpleType()
				.getPrimitive());
	}

	public boolean canBeCastedInto(int castType) {
		return canBeCastedInto(getCastType(this.type), castType);
	}

	public boolean canBeCastedInto(QType castType, byte occurence) {
		if (castType == null)
			return false;
		if (castType instanceof QTypeUnion) {
			QTypeUnion qtypeUnion = (QTypeUnion) castType;
			for (int i = 0; i < qtypeUnion.getList().size(); i++) {
				QType tmpQType = (QType) ((QTypeUnion) castType).getList().get(i);
				byte newOccurence = computeComplexOccurence(occurence, tmpQType.getOccurence());
				if (canBeCastedInto(this.occurence, newOccurence))
					if (canBeCastedInto(tmpQType, newOccurence))
						return true;
			}
		} else if (castType instanceof QTypeSequence) {
			if (((QTypeSequence) castType).getList().size() == 1) {
				QType tmpQType = (QType) ((QTypeSequence) castType).getList().get(0);
				byte newOccurence = computeComplexOccurence(occurence, tmpQType.getOccurence());
				if (canBeCastedInto(this.occurence, newOccurence))
					return canBeCastedInto(tmpQType, newOccurence);
			}
		}
		if (!(castType instanceof QTypeAtom))
			return false;
		return canBeCastedInto(this.occurence, occurence)
				&& canBeCastedInto(getCastType(this.type), getCastType(castType.getType()));
	}

	// clone
	public Object clone() throws CloneNotSupportedException {
		QTypeAtom newObj = new QTypeAtom((SimpleType) type, occurence);
		return newObj;
	}

	// print
	public String toSimpleString() {
		if (type == null)
			return "QTypeAtom(anySimpleType)";
		else
			return "QTypeAtom(" + type.getName() + ")";
	}

	public String toString() {
		return "QTypeAtom(\n" + type + ")";
	}

}
