/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 Odonata.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.xquery.typing;

import org.xquark.schema.*;
import org.xquark.xquery.parser.QName;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryExpression;

public class QTypeElement extends QTypeNode implements QTypeVisitable {
    private static final String RCSRevision = "$Revision: 1.12 $";

    private static final String RCSName = "$Name:  $";

    XQueryExpression name = null;

    ElementDeclaration eltDecl = null;

    public QTypeElement(XQueryExpression name, Type type) {
        this.name = name;
        setType(type);
        subclass = ELEMENT;
        occurence = OCC_1_1;
    }

    public QTypeElement(XQueryExpression name, Type type, byte occurence) {
        this.name = name;
        setType(type);
        subclass = ELEMENT;
        this.occurence = occurence;
    }

    public QTypeElement(XQueryExpression name, ElementDeclaration eltDecl, Type type) {
        this.name = name;
        setType(type);
        setElementDeclaration(eltDecl);
        subclass = ELEMENT;
        occurence = OCC_1_1;
    }

    // public QTypeElement(XQueryExpression name, ElementDeclaration eltDecl, Type type, byte occurence) {
    // this.name = name;
    // setType(type);
    // setElementDeclaration(eltDecl);
    // subclass = ELEMENT;
    // this.occurence = occurence;
    // }

    // #############################################################################
    // VISITOR STUFF
    // #############################################################################

    public void accept(QTypeVisitor visitor) throws XQueryException {
        visitor.visit(this);
    }

    // get/set

    public XQueryExpression getName() {
        return name;
    }

    public ElementDeclaration getElementDeclaration() {
        return eltDecl;
    }

    public void setElementDeclaration(ElementDeclaration eltDecl) {
        this.eltDecl = eltDecl;
    }

    // equals
    public boolean equals(Object qtype) {
        if (!(qtype instanceof QTypeElement))
            return false;
        // TODO verify equality of name
        // if (((QTypeElement) qtype).getOccurence() != occurence)
        // return false;
        return compare(type, ((QTypeElement) qtype).getType()) && name.equals(((QTypeElement) qtype).getName());
    }

    // tells whether qtype is numeric
    public boolean isNumeric() {
        return isNumeric(type);
        /*
         * if (type.isSimpleType()) return isNumeric(type); if (((ComplexType) type).getContentType() == SchemaConstants.MIXED) return true; if (((ComplexType) type).isExtension()) return true; else return false;
         */
    }

    // tells whether qtype is boolean
    public boolean isBoolean() {
        return isBoolean(type);
    }

    // tells whether qtype is integer
    public boolean isInteger() {
        return isInteger(type);
    }

    // tells whether qtype is string
    public boolean isString() {
        if (this.isMixed())
            return true; // this is temporary
        return isString(type);
    }

    // tells whether qtype is an atom
    public boolean isAtom() {
        return false;
    }

    // if (type.isSimpleType()) return true;
    // // else if (((ComplexType) type).getContentType() == SchemaConstants.MIXED) return true;
    // else if (((ComplexType) type).isExtension()) return true;
    // else return false;
    // }

    // tells whether qtype can react to text()
    public boolean isText() {
        return true;
    }

    // tells whether qtype is a date or collection of dates
    public boolean isDate() {
        return isDate(type);
    }

    // tells whether qtype is a datetime or collection of datetimes
    public boolean isDateTime() {
        return isDateTime(type);
    }

    // tells whether qtype is a duration or collection of durations
    public boolean isDuration() {
        return isDuration(type);
    }

    // tells whether qtype is a time or collection of times
    public boolean isTime() {
        return isTime(type);
    }

    // tells whether qtype is a node
    public boolean isNode() {
        return true;
    }

    // tells whether qtype is a node with simple type
    public boolean isSimpleTypeNode() {
        if (type.getValueType() != null)
            return true;
        if (type.getName() != null && type.getName().equals(Type.ANY_TYPE))
            return true;
        return false;
    }

    // tells whether qtype is multiple
    // public boolean isMultiple() {
    // if (type.getName() != null && type.getName().equals("anyType"))
    // return true;
    // return false;
    // }

    // tells whether qtype is atomic (atom or element with simple content
    public boolean isAtomic() {
        if (type.getName() != null && type.getName().equals(TypeVisitor.SC_ANYTYPE))
            return true;
        if (this.isMixed())
            return !isMultiple(); // this is temporary -- add 2002/09/02
        if (type.getValueType() != null)
            return !isMultiple();
        return false;
    }

    // tells whether qtype is mixed
    public boolean isMixed() {
        if (type instanceof ComplexType && ((ComplexType) type).getContentType() == SchemaConstants.MIXED)
            return true;
        return false;
    }

    // tells whether qtype is element
    public boolean isElement() {
        return true;
    }

    public QType applyDATAFunction() throws TypeException {
        if (type.getValueType() != null)
            return new QTypeAtom(type.getValueType(), occurence);
        if (type.getContentType() == Type.MIXED) {
            // SchemaManager sm = new SchemaManager();
            return new QTypeAtom(null, occurence);
        }
        return null;
    }

    // clone
    public Object clone() throws CloneNotSupportedException {
        QTypeElement newObj = new QTypeElement(name, type, occurence);
        return newObj;
    }

    public String toSimpleString() {
        return "QTypeElement(" + name + " ," + type.getName() + ")";
    }

    public String toString() {
        return "QTypeElement(" + name + " ,\n" + type + ")";
    }

    public boolean canBeCastedInto(QType castType, byte occurence) {
        if (castType == null)
            return false;
        if (castType instanceof QTypeUnion) {
            QTypeUnion qtypeUnion = (QTypeUnion) castType;
            for (int i = 0; i < qtypeUnion.getList().size(); i++) {
                QType tmpQType = (QType) ((QTypeUnion) castType).getList().get(i);
                byte newOccurence = computeComplexOccurence(occurence, tmpQType.getOccurence());
                if (canBeCastedInto(this.occurence, newOccurence))
                    if (canBeCastedInto(tmpQType, newOccurence))
                        return true;
            }
        } else if (castType instanceof QTypeSequence) {
            if (((QTypeSequence) castType).getList().size() == 1) {
                QType tmpQType = (QType) ((QTypeSequence) castType).getList().get(0);
                byte newOccurence = computeComplexOccurence(occurence, tmpQType.getOccurence());
                if (canBeCastedInto(this.occurence, newOccurence))
                    return canBeCastedInto(tmpQType, newOccurence);
            }
        }
        if (!(castType instanceof QTypeElement))
            return false;
        return canBeCastedInto(this.occurence, occurence) && canBeCastedInto((QTypeElement) castType);
    }

    private boolean canBeCastedInto(QTypeElement qtype) {
        // TODO check if names and/or element declaration can be casted
        String localName1 = null;
        String namespace1 = null;
        String localName2 = null;
        String namespace2 = null;
        if (this.name instanceof QName) {
            QName qname = (QName) this.name;
            localName1 = qname.getLocalName();
            namespace1 = qname.getNameSpace();
        } else if (eltDecl != null) {
            localName1 = this.eltDecl.getName();
            namespace1 = this.eltDecl.getNamespace();
        }
        if (qtype.getName() instanceof QName) {
            QName qname = (QName) qtype.getName();
            localName2 = qname.getLocalName();
            namespace2 = qname.getNameSpace();
        } else if (qtype.getElementDeclaration() != null) {
            localName2 = qtype.getElementDeclaration().getName();
            namespace2 = qtype.getElementDeclaration().getNamespace();
        }
        if (localName1 == null || localName2 == null)
            return false;
        if (namespace1 != null && (namespace2 == null || !namespace2.equals(namespace1) || !namespace2.equals("*")))
            return false;
        if (!localName2.equals(localName1) && !localName2.equals("*"))
            return false;
        return true;
    }

}
